unit StringTable;

{

  A component that can hold tables of strings, with methods
  to e.g. insert or delete rows and access individual cells.
  It is the successor of the TTableEmulator component, which
  has been declared deprecated.

  Copyright (C) 2016 by FlyingSheep Inc. and Bart Broersma

  This library is free software; you can redistribute it and/or modify it
  under the terms of the GNU Library General Public License as published by
  the Free Software Foundation; either version 2 of the License, or (at your
  option) any later version with the following modification:

  As a special exception, the copyright holders of this library give you
  permission to link this library with independent modules to produce an
  executable, regardless of the license terms of these independent modules,and
  to copy and distribute the resulting executable under terms of your choice,
  provided that you also meet, for each linked independent module, the terms
  and conditions of the license of that module. An independent module is a
  module which is not derived from or based on this library. If you modify
  this library, you may extend this exception to your version of the library,
  but you are not obligated to do so. If you do not wish to do so, delete this
  exception statement from your version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE. See the GNU Library General Public License
  for more details.

  You should have received a copy of the GNU Library General Public License
  along with this library; if not, write to the Free Software Foundation,
  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
}


{$mode objfpc}{$H+}
{$modeswitch typehelpers}

interface

uses
  Classes, SysUtils,
  LazUtf8;

type
  TStringArray = array of String;
  TStringTableSortCompare = function(Item1, Item2: TStringArray): Integer;

  { TStringTable }

  {
    The TStringTable internally works with an array of array of string.
    Since this is a dynamical array of a dynamical array end a managed type
    altering the dimensions of the table is expensive, since it means re-allocation
    of each and every array that gets resized.
    If at al possible, better set RowCount and ColCount at forehand and then fill
    the table using the SetRow function.

    This component was not designed to handle very large tables (> 100 x 100)

    Setting up a 100 x 100 TStringTable on my development machine (jan 2017) took:
    - 1.09 msec using AddRow
    - 0.62 msec using set Rowount/ColCount and then using SetRow

    Setting up a 1000 x 1000 TStringTable on my development machine (jan 2017) took:
    - 104 msec using AddRow
    - 76 msec using set Rowount/ColCount and then using SetRow

    Setting up a 5000 x 5000 TStringTable on my development machine (jan 2017) took:
    - 1966 msec using AddRow
    - 1248 msec using set Rowount/ColCount and then using SetRow

    }

  TStringTable = class
  private
    FAutoGrowColCount: Boolean;
    FRows: array of TStringArray;
    function GetCell(aCol, aRow: Integer): String;
    function GetColCount: Integer;
    function GetRowCount: Integer; inline;
    procedure SetCell(aCol, aRow: Integer; AValue: String);
    procedure SetColCount(AValue: Integer);
    procedure SetRowCount(AValue: Integer);
    procedure MayBeIncreaseColCount(NewCount: Integer);
    procedure QuickSort(L, R: Integer; IsRow: Boolean; CompareFunc: TStringTableSortCompare);
  protected
  public
    constructor Create;
    constructor Create(NrCols, NrRows: Integer); overload;
    destructor Destroy; override;
    function AddRow(Values: array of String): Integer;
    function AddRow(Values: TStrings): Integer;
    function AddCol(Values: array of String): Integer;
    function AddCol(Values: TStrings): Integer;
    procedure DeleteRow(Index: Integer);
    procedure DeleteCol(Index: Integer);
    function InsertRow(Index: Integer; Values: array of String): Integer;
    function InsertRow(Index: Integer; Values: TStrings): Integer;
    function InsertCol(Index: Integer; Values: array of String): Integer;
    function InsertCol(Index: Integer; Values: TStrings): Integer;
    function GetRow(Index: Integer): TStringArray;
    procedure SetRow(Index: Integer; Values: Array of String);
    procedure SetRow(Index: Integer; Values: TStrings);
    function GetCol(Index: Integer): TStringArray;
    procedure SetCol(Index: Integer; Values: Array of String);
    procedure SetCol(Index: Integer; Values: TStrings);

    procedure ExchangeRows(Index1, Index2: Integer);
    procedure ExchangeCols(Index1, Index2: Integer);
    function Find(const S: String; StartRow, StartCol: Integer; CaseSensitive, Partial: Boolean;
                  out ARow, ACol: Integer): Boolean;
    function Find(const S: String;
                  out ARow, ACol: Integer): Boolean;

    procedure Clear;

    procedure Sort(SortRows: Boolean; CompareFunc: TStringTableSortCompare);

    function ToString: String; override;
    function ToString(EscapeMode: TEscapeMode): String; overload;
    procedure DumpTable;

    property AutoGrowColCount: Boolean read FAutoGrowColCount write FAutoGrowColCount default True;
    property Cells[aCol, aRow: Integer]: String read GetCell write SetCell;
    //if you set RowCount or ColCount on an empty table it will affect ColCount resp. RowCount as well
    //if you set either RowCount or ColCount to zero, the StringTable is cleared
    property RowCount: Integer read GetRowCount write SetRowCount;
    property ColCount: Integer read GetColCount write SetColCount;
  end;

{$if FPC_FULLVERSION < 30002}
//fpc <= 3.0.0 requires the declaration to be inside the interface section
//if sourcecode is declared to be UTF8 encoded. This bug is solved in 3.0.2
type
  TStringArrayHelper = type helper for TStringArray
    constructor Create(AValue: TStringArray); overload;
    function Insert(Index: Integer; const S: String): Integer;
    procedure Delete(Index: Integer);
  end;
{$endif}

  EStringTable = class(Exception);

function DbgS(SA: TStringArray; ValuesOnly: Boolean = True): String; overload;
function DbgS(SA: array of string; ValuesOnly: Boolean = True): String; overload;

implementation

const
  SIndexOutOfBounds = 'Index %d out of bounds.';
  SAddRowLengthOutOfBounds = 'Cannot add or set a row with %d values when ColCount = %d';
  SAddColLengthOutOfBounds = 'Cannot add or set a column with %d values when RowCount = %d';
  SCellIndexOutOfBounds = 'Cell index out of bounds: aCol = %d, aRow = %d';


{$if FPC_FULLVERSION >= 30002}
{ TStringArrayHelper }
type
  TStringArrayHelper = type helper for TStringArray
    constructor Create(AValue: TStringArray); overload;
    function Insert(Index: Integer; const S: String): Integer;
    procedure Delete(Index: Integer);
  end;
{$endif}
{ TStringArrayHelper }

constructor TStringArrayHelper.Create(AValue: TStringArray);
begin
  Self := AValue;
end;

function TStringArrayHelper.Insert(Index: Integer; const S: String): Integer;
var
  i: Integer;
begin
  if (Index > (High(Self) + 1)) or (Index < Low(Self)) then
    Raise EStringTable.CreateFmt(SIndexOutOfBounds, [Index]);
  SetLength(Self, Length(Self)+1);
  for i := High(Self) downto (Index + 1) do
    Self[i] := Self[i-1];
  Self[Index] := S;
  Result := Index;
end;

procedure TStringArrayHelper.Delete(Index: Integer);
var
  i: Integer;
begin
  if (Index > High(Self)) or (Index < Low(Self)) then
    Raise EStringTable.CreateFmt(SIndexOutOfBounds, [Index]);
  for i := Index to (High(Self) - 1) do
    Self[i] := Self[i+1];
  SetLength(Self, Length(Self) - 1);
end;

function DbgS(SA: TStringArray; ValuesOnly: Boolean = True): String;
var
  i: Integer;
begin
  if not ValuesOnly then
    Result := Format('Addr: %s, Length: %d, Values: [',[IntToHex({%H-}PtrUInt(Pointer(SA)),2*SizeOf(PtrUInt)),Length(SA)])
  else
    Result := '[';
  for i := Low(SA) to High(SA) do
  begin
    Result := Result + Format('"%s"',[Utf8EscapeControlChars(SA[i], emPascal)]);
    if (i < High(SA)) then
      Result := Result + ',';
  end;
  Result := Result + ']';
end;

function DbgS(SA: array of string; ValuesOnly: Boolean): String;
var
  Tmp: TStringArray;
  i: Integer;
begin
  SetLength(Tmp, Length(SA));
  for i := Low(SA) to High(SA) do
    Tmp[i] := SA[i];
  Result := DbgS(Tmp, ValuesOnly);
  SetLength(Tmp, 0);
end;


{ TStringTable }

function TStringTable.GetCell(aCol, aRow: Integer): String;
begin
  if (aCol > GetColCount - 1) or (aRow > GetRowCount - 1) or
     (aCol < 0) or (aRow < 0) then
     raise EStringTable.CreateFmt(SCellIndexOutOfBounds, [aCol, aRow]);
  //writeln('TStringTable.GetCell: aCol=',acol,', aRow=',arow);
  //writeln('  Length(FRows)=',Length(FRows));
  //if (aRow < Length(FRows)) then writeln('  Length(FRows[ARow])=',Length(FRows[ARow]));
  Result := FRows[ARow][ACol];
end;

function TStringTable.GetColCount: Integer;
begin
  if (Length(FRows) > 0) then
    Result := Length(FRows[0])
  else
    Result := 0;
end;

function TStringTable.GetRowCount: Integer;
begin
  Result := Length(FRows);
end;



procedure TStringTable.SetCell(aCol, aRow: Integer; AValue: String);
begin
  if (aCol > ColCount - 1) or (aRow > RowCount - 1) or
     (aCol < 0) or (aRow < 0) then
     raise EStringTable.CreateFmt(SCellIndexOutOfBounds, [aCol, aRow]);
  FRows[ARow][ACol] := AValue;
end;

procedure TStringTable.SetColCount(AValue: Integer);
var
  i: Integer;
begin
  if (GetColCount = AValue) then Exit;
  if (GetRowCount = 0) then
    SetRowCount(1);
  for i := Low(FRows) to High(FRows) do
  begin
    SetLength(FRows[i], AValue); //this initializes new elements to nil
  end;
end;


procedure TStringTable.SetRowCount(AValue: Integer);
var
  i, OldRowCount: Integer;
begin
  OldRowCount := GetRowCount;
  if (OldRowCount = AValue) then Exit;
  if (AValue = 0) then
  begin
    Clear;
    Exit;
  end;
  if (AValue > OldRowCount) then
  begin//Add rows
    SetLength(FRows, AValue);
    for i := OldRowCount to High(FRows) do
    begin
      if (GetColCount > 0) then
        SetLength(FRows[i], GetColCount) //initialize the row with empty strings
      else
        SetLength(FRows[i], 1); //if table is empty then add rows with 1 column
    end;
  end
  else
  begin
    for i := GetRowCount - 1 downto AValue - 1 do
    begin
      SetLength(FRows[i], 0); //Clear the rows
    end;
    SetLength(FRows, AValue); //Shrink the size
  end;
end;

procedure TStringTable.MayBeIncreaseColCount(NewCount: Integer);
begin
  if (NewCount > GetColCount) then
  //do this also when FRowCount = 0, because we need the adjusted value of ColCount later
  begin
    if FAutoGrowColCount then
    begin
      SetColCount(NewCount); // will create a table with 1 row if table is empty
    end
    else
      Raise EStringTable.CreateFmt(SAddRowLengthOutOfBounds, [NewCount, GetColCount]);
  end;
end;


procedure TStringTable.QuickSort(L, R: Integer; IsRow: Boolean; CompareFunc: TStringTableSortCompare);
var
  Pivot, vL, vR: Integer;
  ExchangeProc: procedure(Index1, Index2: Integer) of object;
  GetFunc: function(Index: Integer): TStringArray of object;
begin
  //if ExchangeItems is override call that, else call (faster) ExchangeItemsInt
  if IsRow then
  begin
    ExchangeProc := @ExchangeRows;
    GetFunc := @GetRow;
  end
  else
  begin
    ExchangeProc := @ExchangeCols;
    GetFunc := @GetCol;
  end;

  if R - L <= 1 then begin // a little bit of time saver
    if L < R then
      if CompareFunc(GetFunc(L), GetFunc(R)) > 0 then
        ExchangeProc(L, R);
    Exit;
  end;

  vL := L;
  vR := R;

  Pivot := L + Random(R - L); // they say random is best

  while vL < vR do begin
    while (vL < Pivot) and (CompareFunc(GetFunc(vL), GetFunc(Pivot)) <= 0) do
      Inc(vL);

    while (vR > Pivot) and (CompareFunc(GetFunc(vR), GetFunc(Pivot)) > 0) do
      Dec(vR);

    ExchangeProc(vL, vR);

    if Pivot = vL then // swap pivot if we just hit it from one side
      Pivot := vR
    else if Pivot = vR then
      Pivot := vL;
  end;

  if Pivot - 1 >= L then
    QuickSort(L, Pivot - 1, IsRow, CompareFunc);
  if Pivot + 1 <= R then
    QuickSort(Pivot + 1, R, IsRow, CompareFunc);
end;

procedure TStringTable.DumpTable;
var
  Len, H, L, i: Integer;
begin
  writeln('============= TStringTable.DumpTable Start ===============');
  Len := Length(FRows);
  H := High(FRows);
  L := Low(FRows);
  writeln('Length(FRows) = ',Len);
  writeln('High(FRows)   = ',H);
  writeln('Low(FRows)    = ',L);
  writeln('RowCount      = ',GetRowCount);
  writeln('ColCount      = ',GetColCount);
  for i := L to H do
  begin
    writeln(i:3,#32,DbgS(FRows[i], False));
  end;
  writeln('============= TStringTable.DumpTable End  ===============');
end;

constructor TStringTable.Create;
begin
  SetLength(FRows, 0);
  FAutoGrowColCount := True;
end;

constructor TStringTable.Create(NrCols, NrRows: Integer);
begin
  Create;
  SetRowCount(NrRows);
  SetColCount(NrCols);
end;

destructor TStringTable.Destroy;
begin
  Clear;
  SetLength(FRows, 0);
  inherited Destroy;
end;

function TStringTable.AddRow(Values: array of String): Integer;
begin
  Result := InsertRow(GetRowCount, Values);
end;

function TStringTable.AddRow(Values: TStrings): Integer;
var
  Tmp: TStringArray;
  i: Integer;
begin
  //writeln('TStringTable.AddRow (TStrings)');
  SetLength(Tmp, Values.Count);
  for i := 0 to Values.Count - 1 do
    Tmp[i] := Values.Strings[i];
  Result := AddRow(Tmp);
  SetLength(Tmp, 0);
end;

function TStringTable.AddCol(Values: array of String): Integer;
begin
  Result := InsertCol(GetColCount, Values);
end;

function TStringTable.AddCol(Values: TStrings): Integer;
var
  Tmp: TStringArray;
  i: Integer;
begin
  SetLength(Tmp, Values.Count);
  for i := 0 to Values.Count - 1 do
    Tmp[i] := Values.Strings[i];
  Result := AddCol(Tmp);
  SetLength(Tmp, 0);
end;

{
Adapted from
http://stackoverflow.com/questions/3730709/move-to-insert-delete-items-from-a-dynamic-array-of-string
}
procedure TStringTable.DeleteRow(Index: Integer);
var
  Len, TailElements: Integer;
begin
  if (Index > GetRowCount - 1) or (Index < 0) then
    Raise EStringTable.CreateFmt(SIndexOutOfBounds, [Index]);
  if (GetRowCount = 1) then
    Clear
  else
  begin
    Len := Length(FRows);
    Finalize(FRows[Index]);
    //SetLength(FRows[Index], 0);
    //for i := Index to High(FRows) - 1 do
    //  FRows[i] := FRows[i+1];
    TailElements := Len - Index;
    if TailElements > 0 then
      Move(FRows[Index + 1], FRows[Index], SizeOf(TStringArray) * TailElements);
    Initialize(FRows[Len - 1]);
    SetLength(FRows, Len - 1);
    //SetLength(FRows, GetRowCount-1);
  end;
end;

procedure TStringTable.DeleteCol(Index: Integer);
var
  i: Integer;
begin
  if (Index > GetColCount - 1) or (Index < 0) then
    Raise EStringTable.CreateFmt(SIndexOutOfBounds, [Index]);
  if (GetColCount = 1) then
    Clear
  else
  begin
    for i := Low(FRows) to High(FRows) do
      FRows[i].Delete(Index);
  end;
end;


{
Adapted from
http://stackoverflow.com/questions/3730709/move-to-insert-delete-items-from-a-dynamic-array-of-string
}
function TStringTable.InsertRow(Index: Integer; Values: array of String): Integer;
var
  i, Len, TailElements: Integer;
  TableWasEmpty: Boolean;
begin
  //writeln('InsertRow: Index = ',Index,', GetRowCount = ',GetRowCount);
  if (Index < 0) or (Index > GetRowCount) then
    Raise EStringTable.CreateFmt(SIndexOutOfBounds, [Index]);
  Len := Length(FRows);
  TableWasEmpty := (GetRowCount = 0);
  MayBeIncreaseColCount(Length(Values));
  if (not TableWasEmpty) then //table was empty before, call to MayBeIncreaseColCount already created 1 row!
  begin
    SetLength(FRows, Len + 1);
    Finalize(FRows[Len]);
  end;
  TailElements := Len - Index;
  //SetLength(FRows[Index], GetColCount);
  //for i := High(FRows) downto (Index + 1) do
  //  FRows[i] := FRows[i-1];
  if (TailElements > 0) then
    Move(FRows[Index], FRows[Index + 1], SizeOf(TStringArray) * TailElements);
  if not TableWasEmpty then
  begin
    Initialize(FRows[Index]);
    SetLength(FRows[Index], GetColCount);
  end;
  for i := Low(Values) to High(Values) do
    FRows[Index][i] := Values[i];
  //FRows[Index] := Values;

  Result := Index;
end;

function TStringTable.InsertRow(Index: Integer; Values: TStrings): Integer;
var
  Tmp: TStringArray;
  i: Integer;
begin
  SetLength(Tmp, Values.Count);
  for i := 0 to Values.Count - 1 do
    Tmp[i] := Values.Strings[i];
  Result := InsertRow(Index, Tmp);
  SetLength(Tmp, 0);
end;

function TStringTable.InsertCol(Index: Integer; Values: array of String): Integer;
var
  i: Integer;
begin
  if (GetRowCount = 0) and (Index = 0) and (Length(Values) = 1) then
  begin//Allow inserting a single 1-value column into an empty table
    Result := InsertRow(0, Values);
    Exit;
  end;
  if (Index > GetColCount) or (Index < 0) then
    Raise EStringTable.CreateFmt(SIndexOutOfBounds, [Index]);
  if (Length(Values) > GetRowCount) then
    Raise EStringTable.CreateFmt(SAddColLengthOutOfBounds, [Length(Values), GetRowCount]);
  for i := Low(FRows) to High(FRows) do
  begin
    if (i <= High(Values)) then
      FRows[i].Insert(Index, Values[i])
    else
      FRows[i].Insert(Index, '');
  end;
  Result := Index;
end;

function TStringTable.InsertCol(Index: Integer; Values: TStrings): Integer;
var
  Tmp: TStringArray;
  i: Integer;
begin
  SetLength(Tmp, Values.Count);
  for i := 0 to Values.Count - 1 do
    Tmp[i] := Values.Strings[i];
  Result := InsertCol(Index, Tmp);
  SetLength(Tmp, 0);
end;

function TStringTable.GetRow(Index: Integer): TStringArray;
var
  i: Integer;
begin
  if (Index < 0) or (Index > GetRowCount - 1) then
    Raise EStringTable.CreateFmt(SIndexOutOfBounds, [Index]);
  SetLength(Result, Length(FRows[Index]));
  for i := Low(FRows[Index]) to High(FRows[Index]) do
    Result[i] := FRows[index][i];
end;

procedure TStringTable.SetRow(Index: Integer; Values: Array of String);
var
  i: Integer;
begin
  if (Index < 0) or (Index > GetRowCount - 1) then
    Raise EStringTable.CreateFmt(SIndexOutOfBounds, [Index]);
  MayBeIncreaseColCount(Length(Values));
  for i := Low(FRows[Index]) to High(FRows[Index]) do
  begin
    if (i <= High(Values)) then
      FRows[Index][i] := Values[i]
     else
       FRows[Index][i] := '';
  end;
end;

procedure TStringTable.SetRow(Index: Integer; Values: TStrings);
var
  Tmp: TStringArray;
  i: Integer;
begin
  SetLength(Tmp, Values.Count);
  for i := 0 to Values.Count - 1 do
    Tmp[i] := Values.Strings[i];
  SetRow(Index, Tmp);
  SetLength(Tmp, 0);
end;

function TStringTable.GetCol(Index: Integer): TStringArray;
var
  i: Integer;
begin
  if (Index < 0) or (Index > GetColCount - 1) then
    Raise EStringTable.CreateFmt(SIndexOutOfBounds, [Index]);
  SetLength(Result, GetRowCount);
  for i := Low(FRows) to High(FRows) do
    Result[i] := FRows[i][Index];
end;

procedure TStringTable.SetCol(Index: Integer; Values: array of String);
var
  i: Integer;
begin
  if (GetRowCount = 0) or (Index > GetColCount) or (Index < 0) then
    Raise EStringTable.CreateFmt(SIndexOutOfBounds, [Index]);
  if (Length(Values) > GetRowCount) then
    Raise EStringTable.CreateFmt(SAddColLengthOutOfBounds, [Length(Values), GetRowCount]);
  for i := Low(FRows) to High(FRows) do
  begin
    if (i <= High(Values)) then
      FRows[i][Index] := Values[i]
    else
      FRows[i][Index] := '';
  end;
end;

procedure TStringTable.SetCol(Index: Integer; Values: TStrings);
var
  Tmp: TStringArray;
  i: Integer;
begin
  SetLength(Tmp, Values.Count);
  for i := 0 to Values.Count - 1 do
    Tmp[i] := Values.Strings[i];
  SetCol(Index, Tmp);
  SetLength(Tmp, 0);
end;

procedure TStringTable.ExchangeRows(Index1, Index2: Integer);
var
  Tmp: TStringArray;
begin
  if (Index1 < 0) or (Index1 > GetRowCount - 1) then
    Raise EStringTable.CreateFmt(SIndexOutOfBounds, [Index1]);
  if (Index2 < 0) or (Index2 > GetRowCount - 1) then
    Raise EStringTable.CreateFmt(SIndexOutOfBounds, [Index2]);
  if (Index1 = Index2) then
    Exit;
  Tmp := FRows[Index1];
  FRows[Index1] := FRows[Index2];
  FRows[Index2] := Tmp;
end;

procedure TStringTable.ExchangeCols(Index1, Index2: Integer);
var
  Tmp: String;
  i: Integer;
begin
  if (Index1 < 0) or (Index1 > GetColCount - 1) then
    Raise EStringTable.CreateFmt(SIndexOutOfBounds, [Index1]);
  if (Index2 < 0) or (Index2 > GetColCount - 1) then
    Raise EStringTable.CreateFmt(SIndexOutOfBounds, [Index2]);
  if (Index1 = Index2) then
    Exit;
  for i := Low(FRows) to High(FRows) do
  begin
    Tmp := FRows[i][Index1];
    FRows[i][Index1] := FRows[i][Index2];
    FRows[i][Index2] := Tmp;
  end;
end;

function TStringTable.Find(const S: String; StartRow, StartCol: Integer;
                           CaseSensitive, Partial: Boolean;
                           out ARow, ACol: Integer): Boolean;
var
  CellText: String;
begin
  Result := False;
  if (StartRow < 0) or (StartRow > GetRowCount - 1) then
    Raise EStringTable.CreateFmt(SIndexOutOfBounds, [StartRow]);
  if (StartCol < 0) or (StartCol > GetColCount - 1) then
    Raise EStringTable.CreateFmt(SIndexOutOfBounds, [StartCol]);
  ARow := StartRow;
  ACol := StartCol;
  repeat
    repeat
      CellText := FRows[ARow][ACol];
      if not Partial then
      begin //not Partial
        if CaseSensitive then
          Result := (S = CellText)
        else
          Result := (Utf8CompareText(S, CellText) = 0);
      end
      else
      begin //Partial
        if CaseSensitive then
          Result := (Utf8Pos(S, CellText) > 0)
        else
          Result := (Utf8Pos(Utf8UpperCase(S), Utf8UpperCase(CellText)) > 0)
      end;
      if not Result then
        Inc(ACol);
    until Result or (ACol > GetColCount - 1);
    if not Result then
    begin
      Inc(ARow);
      ACol := 0;
    end;
  until Result or (ARow > GetRowCount - 1);
end;

function TStringTable.Find(const S: String; out ARow, ACol: Integer): Boolean;
begin
  Result := Find(S, 0, 0, False, False, ARow, ACol);
end;

procedure TStringTable.Clear;
var
  i: Integer;
begin
  for i := Low(FRows) to High(FRows) do
    SetLength(FRows[i], 0);
  SetLength(FRows, 0);
end;

procedure TStringTable.Sort(SortRows: Boolean;
  CompareFunc: TStringTableSortCompare);
begin
  if SortRows then
  begin
    if (GetRowCount > 1) then
      QuickSort(0, GetRowCount-1, SortRows, CompareFunc)
  end
  else
  begin
    if (GetColCount > 1) then
      QuickSort(0, GetColCount-1, SortRows, CompareFunc);
  end;
end;

function TStringTable.ToString: String;
var
  R, C: Integer;
begin
  Result := '';
  for R := 0 to RowCount - 1 do
  begin
    Result := Result + '|';
    for C := 0 to GetColCount - 1 do Result := Result + Cells[C,R] + '|';
    if (R < GetRowCount - 1) then Result := Result + LineEnding;
  end;
end;

function TStringTable.ToString(EscapeMode: TEscapeMode): String;
var
  R, C: Integer;
begin
  Result := '';
  for R := 0 to RowCount - 1 do
  begin
    Result := Result + '|';
    for C := 0 to GetColCount - 1 do Result := Result + Utf8EscapeControlChars(Cells[C,R], EscapeMode) + '|';
    if (R < GetRowCount - 1) then Result := Result + LineEnding;
  end;
end;

end.

